#!/usr/bin/env node

import { execSync } from "node:child_process";
import { cp, mkdir, mkdtemp, readdir, rm } from "node:fs/promises";
import { createServer } from "node:net";
import { tmpdir } from "node:os";
import { join } from "node:path";
import { stdin as input, stdout as output } from "node:process";
import { createInterface } from "node:readline/promises";
import { cloneRepo, removeGit, tryGitInit } from "./git.mjs";

try {
  await Promise.all([isPortTaken(5432), isPortTaken(5433)]);
} catch (e) {
  console.log(e.message);
  process.exit(1);
}

const rl = createInterface({ input, output });

let workingDir = "";
let isRemoveDocker = false;

{
  const answer = await rl.question(question("What is your project named?"));

  workingDir = join(process.cwd(), answer);

  try {
    await mkdir(workingDir);
  } catch {
    error("The directory already exists");
    process.exit(1);
  }
}
{
  const answer = await rl.question(
    question("Do you want to remove Docker for app? (y/N)"),
  );

  if (answer === "y" || answer === "Y") {
    isRemoveDocker = true;
  }
}

rl.close();

process.chdir(workingDir);

if (process.env.LOCAL_FROM_PATH) {
  await copyDir(process.env.LOCAL_FROM_PATH, workingDir);
} else {
  cloneRepo();
  report("Completed to clone hiroppy/web-app-template");
}

removeGit(workingDir);
tryGitInit(workingDir);

execSync("npm run setup", { stdio: "ignore" });

report("Installing dependencies...");
execSync("pnpm i", { stdio: "ignore" });

execSync("cp .env.sample .env.local");

report("Setting up...");

if (isRemoveDocker) {
  execSync("node .internal/init.mjs --skip-questions --remove-docker", {
    stdio: "ignore",
  });
} else {
  execSync("node .internal/init.mjs --skip-questions", { stdio: "ignore" });
}

console.log("");
console.log("Completed to setup 🎉🎉🎉🎉🎉🎉");
console.log("");
console.log(
  "🖼️  this code base is https://github.com/hiroppy/web-app-template.",
);

function question(title) {
  return `🎃 ${title}  `;
}

function report(text) {
  console.log("\x1b[36m%s\x1b[0m", `🐕  ${text}`);
}

function error(text) {
  console.error("\x1b[31m%s\x1b[0m", `🔥  ${text} (exited)`);
}

async function isPortTaken(port) {
  return new Promise((resolve, reject) => {
    const tester = createServer()
      .once("error", (err) => {
        reject(err);
      })
      .once("listening", () => {
        tester.once("close", () => resolve(true)).close();
      })
      .listen(port);
  });
}

async function copyDir(from, to) {
  console.log(`copy from ${process.env.LOCAL_FROM_PATH}`);
  console.log(`copy to ${to}`);

  const tmpDir = join(tmpdir(), "create-app-foundation-");

  console.log(`tmpDir: ${tmpDir}`);

  await copyDirectory(from, tmpDir, [
    // add files written in .gitignore
    "node_modules",
    ".next",
    "test-results",
    ".env.local",
    "next-env.d.ts",
    "prisma/migrations",
  ]);
  await cp(tmpDir, to, { recursive: true });
  await rm(tmpDir, { recursive: true });
}

async function copyDirectory(src, dest, exclude = []) {
  await mkdir(dest, { recursive: true });

  const entries = await readdir(src, { withFileTypes: true });

  for (const entry of entries) {
    const srcPath = join(src, entry.name);
    const destPath = join(dest, entry.name);

    if (exclude.some((excluded) => srcPath.includes(excluded))) {
      console.log(`Excluding: ${srcPath}`);
      continue;
    }

    if (entry.isDirectory()) {
      await copyDirectory(srcPath, destPath, exclude);
    } else {
      await cp(srcPath, destPath);
    }
  }
}
